From 0fd79a1b08fe7c7c7a2e75b3f7bf3e7dad05dc9f Mon Sep 17 00:00:00 2001
From: Max Lapshin <max@maxidoors.ru>
Date: Wed, 21 Oct 2009 14:13:17 +0400
Subject: [PATCH] capture stderr of spawned process

---
 erts/emulator/beam/erl_bif_port.c |   11 ++-
 erts/emulator/beam/erl_driver.h   |    1 +
 erts/emulator/beam/sys.h          |    2 +-
 erts/emulator/sys/unix/sys.c      |  183 ++++++++++++++++++++++++++-----------
 4 files changed, 139 insertions(+), 58 deletions(-)

diff --git a/erts/emulator/beam/erl_bif_port.c b/erts/emulator/beam/erl_bif_port.c
index 7ac236e..f961b18 100644
--- a/erts/emulator/beam/erl_bif_port.c
+++ b/erts/emulator/beam/erl_bif_port.c
@@ -693,6 +693,8 @@ open_port(Process* p, Eterm name, Eterm settings, int *err_nump)
 		opts.use_stdio = 1;
 	    } else if (*nargs == am_stderr_to_stdout) {
 		opts.redir_stderr = 1;
+	    } else if (*nargs == am_error) {
+		opts.read_write |= DO_ERROR;
 	    } else if (*nargs == am_line) {
 		linebuf = 512;
 	    } else if (*nargs == am_nouse_stdio) {
@@ -720,15 +722,16 @@ open_port(Process* p, Eterm name, Eterm settings, int *err_nump)
 	    nargs = list_val(*nargs);
 	}
     }
-    if (opts.read_write == 0)	/* implement default */
-	opts.read_write = DO_READ|DO_WRITE;
+    if (!(opts.read_write & DO_READ) && !(opts.read_write & DO_WRITE))	/* implement default */
+	opts.read_write |= DO_READ|DO_WRITE;
 
     /* Mutually exclusive arguments. */
     if((linebuf && opts.packet_bytes) || 
-       (opts.redir_stderr && !opts.use_stdio)) {
+       (opts.redir_stderr && !opts.use_stdio) ||
+       (opts.redir_stderr && (opts.read_write & DO_ERROR))) {
 	goto badarg;
     }
-
+     
     /*
      * Parse the first argument and start the appropriate driver.
      */
diff --git a/erts/emulator/beam/erl_driver.h b/erts/emulator/beam/erl_driver.h
index 6cc2365..767448f 100644
--- a/erts/emulator/beam/erl_driver.h
+++ b/erts/emulator/beam/erl_driver.h
@@ -76,6 +76,7 @@ typedef struct {
 /* Old deprecated */
 #define DO_READ  ERL_DRV_READ
 #define DO_WRITE ERL_DRV_WRITE
+#define DO_ERROR (1 << 4)
 
 #define ERL_DRV_EXTENDED_MARKER		(0xfeeeeeed)
 #define ERL_DRV_EXTENDED_MAJOR_VERSION	1
diff --git a/erts/emulator/beam/sys.h b/erts/emulator/beam/sys.h
index 9346b8f..efd32e5 100644
--- a/erts/emulator/beam/sys.h
+++ b/erts/emulator/beam/sys.h
@@ -538,7 +538,7 @@ typedef struct _SysDriverOpts {
     int ifd;			/* Input file descriptor (fd driver). */
     int ofd;			/* Outputfile descriptor (fd driver). */
     int packet_bytes;		/* Number of bytes in packet header. */
-    int read_write;		/* Read and write bits. */
+    int read_write;		/* Read, write and catch stderr bits. */
     int use_stdio;		/* Use standard I/O: TRUE or FALSE. */
     int redir_stderr;           /* Redirect stderr to stdout: TRUE/FALSE. */
     int hide_window;		/* Hide this windows (Windows). */
diff --git a/erts/emulator/sys/unix/sys.c b/erts/emulator/sys/unix/sys.c
index 183525b..ccd9889 100644
--- a/erts/emulator/sys/unix/sys.c
+++ b/erts/emulator/sys/unix/sys.c
@@ -111,6 +111,7 @@ struct ErtsSysReportExit_ {
     int pid;
     int ifd;
     int ofd;
+    int efd;
 #if CHLDWTHR && !defined(ERTS_SMP)
     int status;
 #endif
@@ -1022,11 +1023,12 @@ void fini_getenv_state(GETENV_STATE *state)
 
 /* This data is shared by these drivers - initialized by spawn_init() */
 static struct driver_data {
-    int port_num, ofd, packet_bytes;
+    int port_num, ofd, efd, packet_bytes;
     ErtsSysReportExit *report_exit;
     int pid;
     int alive;
     int status;
+		int error_stream; /* is this fd a stderr of spawned process */
 } *driver_data;			/* indexed by fd */
 
 /* Driver interfaces */
@@ -1163,6 +1165,7 @@ static RETSIGTYPE onchld(int signum)
 static int set_driver_data(int port_num,
 			   int ifd,
 			   int ofd,
+			   int efd,
 			   int packet_bytes,
 			   int read_write,
 			   int exit_status,
@@ -1180,12 +1183,26 @@ static int set_driver_data(int port_num,
 	report_exit->pid = pid;
 	report_exit->ifd = read_write & DO_READ ? ifd : -1;
 	report_exit->ofd = read_write & DO_WRITE ? ofd : -1;
+	report_exit->efd = read_write & DO_ERROR ? efd : -1;
 #if CHLDWTHR && !defined(ERTS_SMP)
 	report_exit->status = 0;
 #endif
 	report_exit_list = report_exit;
     }
 
+		if (read_write & DO_ERROR) {
+			driver_data[efd].packet_bytes = packet_bytes;
+			driver_data[efd].port_num = port_num;
+			driver_data[efd].report_exit = report_exit;
+			driver_data[efd].pid = pid;
+			driver_data[efd].alive = 1;
+			driver_data[efd].status = 0;
+			driver_data[efd].ofd = -1;
+			driver_data[efd].efd = -1;
+			driver_data[efd].error_stream = 1;
+			(void) driver_select(port_num, efd, (ERL_DRV_READ|ERL_DRV_USE), 1);
+		}
+
     if (read_write & DO_READ) {
 	driver_data[ifd].packet_bytes = packet_bytes;
 	driver_data[ifd].port_num = port_num;
@@ -1193,6 +1210,10 @@ static int set_driver_data(int port_num,
 	driver_data[ifd].pid = pid;
 	driver_data[ifd].alive = 1;
 	driver_data[ifd].status = 0;
+	if (read_write & DO_ERROR) {
+		driver_data[ifd].efd = efd;
+	} else
+	  driver_data[ifd].efd = -1;
 	if (read_write & DO_WRITE) {
 	    driver_data[ifd].ofd = ofd;
 	    if (ifd != ofd)
@@ -1201,6 +1222,7 @@ static int set_driver_data(int port_num,
 	    driver_data[ifd].ofd = -1;
 	}
 	(void) driver_select(port_num, ifd, (ERL_DRV_READ|ERL_DRV_USE), 1);
+	
 	return(ifd);
     } else {			/* DO_WRITE only */
 	driver_data[ofd].packet_bytes = packet_bytes;
@@ -1210,6 +1232,10 @@ static int set_driver_data(int port_num,
 	driver_data[ofd].pid = pid;
 	driver_data[ofd].alive = 1;
 	driver_data[ofd].status = 0;
+	if (read_write & DO_ERROR) {
+		driver_data[ifd].efd = efd;
+	} else
+	  driver_data[ifd].efd = -1;
 	return(ofd);
     }
 }
@@ -1245,7 +1271,7 @@ static int spawn_init()
    return 1;
 }
 
-static void close_pipes(int ifd[2], int ofd[2], int read_write)
+static void close_pipes(int ifd[2], int ofd[2], int efd[2], int read_write)
 {
     if (read_write & DO_READ) {
 	(void) close(ifd[0]);
@@ -1255,6 +1281,10 @@ static void close_pipes(int ifd[2], int ofd[2], int read_write)
 	(void) close(ofd[0]);
 	(void) close(ofd[1]);
     }
+    if (read_write & DO_ERROR) {
+	(void) close(efd[0]);
+	(void) close(efd[1]);
+    }
 }
 
 static void init_fd_data(int fd, int prt)
@@ -1354,7 +1384,7 @@ static ErlDrvData spawn_start(ErlDrvPort port_num, char* name, SysDriverOpts* op
 #define CMD_LINE_PREFIX_STR "exec "
 #define CMD_LINE_PREFIX_STR_SZ (sizeof(CMD_LINE_PREFIX_STR) - 1)
 
-    int ifd[2], ofd[2], len, pid, i;
+    int ifd[2], ofd[2], efd[2], len, pid, i;
     char **volatile new_environ; /* volatile since a vfork() then cannot
 				    cause 'new_environ' to be clobbered
 				    in the parent process. */
@@ -1373,50 +1403,49 @@ static ErlDrvData spawn_start(ErlDrvPort port_num, char* name, SysDriverOpts* op
 				&no_vfork_sz) >= 0);
 #endif
 
-    switch (opts->read_write) {
-    case DO_READ:
-	if (pipe(ifd) < 0)
-	    return ERL_DRV_ERROR_ERRNO;
-	if (ifd[0] >= max_files) {
-	    close_pipes(ifd, ofd, opts->read_write);
-	    errno = EMFILE;
-	    return ERL_DRV_ERROR_ERRNO;
-	}
-	ofd[1] = -1;		/* keep purify happy */
-	break;
-    case DO_WRITE:
-	if (pipe(ofd) < 0) return ERL_DRV_ERROR_ERRNO;
-	if (ofd[1] >= max_files) {
-	    close_pipes(ifd, ofd, opts->read_write);
-	    errno = EMFILE;
-	    return ERL_DRV_ERROR_ERRNO;
-	}
-	ifd[0] = -1;		/* keep purify happy */
-	break;
-    case DO_READ|DO_WRITE:
-	if (pipe(ifd) < 0) return ERL_DRV_ERROR_ERRNO;
-	errno = EMFILE;		/* default for next two conditions */
-	if (ifd[0] >= max_files || pipe(ofd) < 0) {
-	    close_pipes(ifd, ofd, DO_READ);
-	    return ERL_DRV_ERROR_ERRNO;
-	}
-	if (ofd[1] >= max_files) {
-	    close_pipes(ifd, ofd, opts->read_write);
-	    errno = EMFILE;
-	    return ERL_DRV_ERROR_ERRNO;
-	}
-	break;
-    default:
-	ASSERT(0);
-	return ERL_DRV_ERROR_GENERAL;
-    }
+	  ifd[0] = -1;		/* keep purify happy */
+	  ifd[1] = -1;		/* keep purify happy */
+	  ofd[0] = -1;		/* keep purify happy */
+	  ofd[1] = -1;		/* keep purify happy */
+		efd[0] = -1;
+		efd[1] = -1;
+
+    if (opts->read_write & DO_READ) {
+	  	if (pipe(ifd) < 0)
+	  	    return ERL_DRV_ERROR_ERRNO;
+	  	if (ifd[0] >= max_files) {
+	  	    close_pipes(ifd, ofd, efd, opts->read_write);
+	  	    errno = EMFILE;
+	  	    return ERL_DRV_ERROR_ERRNO;
+	  	}
+		}
+		if (opts->read_write & DO_WRITE) {
+	  	if (pipe(ofd) < 0) return ERL_DRV_ERROR_ERRNO;
+	  	if (ofd[1] >= max_files) {
+	  	    close_pipes(ifd, ofd, efd, opts->read_write);
+	  	    errno = EMFILE;
+	  	    return ERL_DRV_ERROR_ERRNO;
+	  	}
+		}
+		if (opts->read_write & DO_ERROR) {
+	  	if (pipe(efd) < 0) return ERL_DRV_ERROR_ERRNO;
+	  	if (efd[0] >= max_files) {
+	  	    close_pipes(ifd, ofd, efd, opts->read_write);
+	  	    errno = EMFILE;
+	  	    return ERL_DRV_ERROR_ERRNO;
+	  	}
+		}
+		if (opts->read_write & (!(DO_READ | DO_WRITE | DO_ERROR))) {
+			ASSERT(0);
+			return ERL_DRV_ERROR_GENERAL;
+		}
 
     if (opts->spawn_type == ERTS_SPAWN_EXECUTABLE) {
 	/* started with spawn_executable, not with spawn */
 	len = strlen(name);
 	cmd_line = (char *) erts_alloc_fnf(ERTS_ALC_T_TMP, len + 1);
 	if (!cmd_line) {
-	    close_pipes(ifd, ofd, opts->read_write);
+	    close_pipes(ifd, ofd, efd, opts->read_write);
 	    errno = ENOMEM;
 	    return ERL_DRV_ERROR_ERRNO;
 	}
@@ -1434,7 +1463,7 @@ static ErlDrvData spawn_start(ErlDrvPort port_num, char* name, SysDriverOpts* op
 	cmd_line = (char *) erts_alloc_fnf(ERTS_ALC_T_TMP,
 					   CMD_LINE_PREFIX_STR_SZ + len + 1);
 	if (!cmd_line) {
-	    close_pipes(ifd, ofd, opts->read_write);
+	    close_pipes(ifd, ofd, efd, opts->read_write);
 	    errno = ENOMEM;
 	    return ERL_DRV_ERROR_ERRNO;
 	}
@@ -1446,7 +1475,6 @@ static ErlDrvData spawn_start(ErlDrvPort port_num, char* name, SysDriverOpts* op
     }
 
     erts_smp_rwmtx_rlock(&environ_rwmtx);
-
     if (opts->envir == NULL) {
 	new_environ = environ;
     } else if ((new_environ = build_unix_environment(opts->envir)) == NULL) {
@@ -1486,6 +1514,12 @@ static ErlDrvData spawn_start(ErlDrvPort port_num, char* name, SysDriverOpts* op
 	     * erl_child_setup.c) up to date if changes are made here.
 	     */
 
+	    if(opts->read_write & DO_ERROR) {
+		/* stderr for process, will not be mixed with opts->redir_stderr because of parameter check */
+		if (dup2(efd[1], 2) < 0)
+		    goto child_error;
+     }
+
 	    if (opts->use_stdio) {
 		if (opts->read_write & DO_READ) {
 		    /* stdout for process */
@@ -1556,6 +1590,9 @@ static ErlDrvData spawn_start(ErlDrvPort port_num, char* name, SysDriverOpts* op
 	/* Setup argv[] for the child setup program (implemented in
 	   erl_child_setup.c) */
 	i = 0;
+	if (opts->read_write & DO_ERROR) {
+		sprintf(&dup2_op[i++][0], "%d:%d", efd[1], 2);
+	}
 	if (opts->use_stdio) {
 	    if (opts->read_write & DO_READ){
 		/* stdout for process */
@@ -1638,7 +1675,7 @@ static ErlDrvData spawn_start(ErlDrvPort port_num, char* name, SysDriverOpts* op
 	erts_smp_rwmtx_runlock(&environ_rwmtx);
 	erts_free(ERTS_ALC_T_TMP, (void *) cmd_line);
         unblock_signals();
-        close_pipes(ifd, ofd, opts->read_write);
+        close_pipes(ifd, ofd, efd, opts->read_write);
 	errno = saved_errno;
 	return ERL_DRV_ERROR_ERRNO;
     }
@@ -1665,7 +1702,7 @@ static ErlDrvData spawn_start(ErlDrvPort port_num, char* name, SysDriverOpts* op
 	erts_free(ERTS_ALC_T_TMP, (void *) cmd_line);
         reset_qnx_spawn();
 	erts_smp_rwmtx_runlock(&environ_rwmtx);
-	close_pipes(ifd, ofd, opts->read_write);
+	close_pipes(ifd, ofd, efd, opts->read_write);
 	return ERL_DRV_ERROR_GENERAL;
     }
     reset_qnx_spawn();
@@ -1680,6 +1717,8 @@ static ErlDrvData spawn_start(ErlDrvPort port_num, char* name, SysDriverOpts* op
 	(void) close(ifd[1]);
     if (opts->read_write & DO_WRITE)
 	(void) close(ofd[0]);
+	  if (opts->read_write & DO_ERROR)
+	(void) close(efd[1]);
 	
     if (opts->read_write & DO_READ) {
 	SET_NONBLOCKING(ifd[0]);
@@ -1689,8 +1728,12 @@ static ErlDrvData spawn_start(ErlDrvPort port_num, char* name, SysDriverOpts* op
 	SET_NONBLOCKING(ofd[1]);
         init_fd_data(ofd[1], port_num);
     }
+    if (opts->read_write & DO_ERROR) {
+	SET_NONBLOCKING(efd[0]);
+        init_fd_data(efd[0], port_num);
+    }
 
-    res = set_driver_data(port_num, ifd[0], ofd[1], opts->packet_bytes,
+    res = set_driver_data(port_num, ifd[0], ofd[1], efd[0], opts->packet_bytes,
 			  opts->read_write, opts->exit_status, pid);
     /* Don't unblock SIGCHLD until now, since the call above must
        first complete putting away the info about our new subprocess. */
@@ -1919,7 +1962,7 @@ static ErlDrvData fd_start(ErlDrvPort port_num, char* name,
 	}
     }
     CHLD_STAT_LOCK;
-    res = (ErlDrvData)(long)set_driver_data(port_num, opts->ifd, opts->ofd,
+    res = (ErlDrvData)(long)set_driver_data(port_num, opts->ifd, opts->ofd, -1,
 				      opts->packet_bytes,
 				      opts->read_write, 0, -1);
     CHLD_STAT_UNLOCK;
@@ -1976,7 +2019,7 @@ static ErlDrvData vanilla_start(ErlDrvPort port_num, char* name,
     init_fd_data(fd, port_num);
 
     CHLD_STAT_LOCK;
-    res = (ErlDrvData)(long)set_driver_data(port_num, fd, fd,
+    res = (ErlDrvData)(long)set_driver_data(port_num, fd, fd, -1, 
 				      opts->packet_bytes,
 				      opts->read_write, 0, -1);
     CHLD_STAT_UNLOCK;
@@ -1988,7 +2031,7 @@ static ErlDrvData vanilla_start(ErlDrvPort port_num, char* name,
 
 static void stop(ErlDrvData fd)
 {
-    int prt, ofd;
+    int prt, ofd, efd;
 
     prt = driver_data[(int)(long)fd].port_num;
     nbio_stop_fd(prt, (int)(long)fd);
@@ -1998,6 +2041,12 @@ static void stop(ErlDrvData fd)
 	nbio_stop_fd(prt, ofd);
     else
 	ofd = -1;
+	
+	  efd = driver_data[(int)(long)fd].efd;
+	  if (efd != (int)(long)fd && (int)(long)efd != -1)
+	nbio_stop_fd(prt, efd);
+	  else
+	efd = -1;
 
     CHLD_STAT_LOCK;
 
@@ -2012,6 +2061,9 @@ static void stop(ErlDrvData fd)
     if (ofd >= 0) {
 	driver_select(prt, (int)(long)ofd, ERL_DRV_USE, 0);  /* close(ofd); */
     }
+    if (efd >= 0) {
+	driver_select(prt, (int)(long)efd, ERL_DRV_USE, 0);  /* close(efd); */
+    }
 }
 
 static void outputv(ErlDrvData e, ErlIOVec* ev)
@@ -2163,6 +2215,23 @@ static int port_inp_failure(int port_num, int ready_fd, int res)
 /* initial start routine                        */
 /* ready_fd is the descriptor that is ready to read */
 
+
+static void send_from_spawned_process(int fd, char *buf, int size)
+{
+	if (!driver_data[fd].error_stream)
+	  driver_output(driver_data[fd].port_num, buf, size);
+	else {
+		ErlDrvTermData spec[] = {
+			ERL_DRV_ATOM, am_error,
+			ERL_DRV_BUF2BINARY, buf, size,
+			ERL_DRV_TUPLE, 2
+		};
+		
+   	driver_output_term(driver_data[fd].port_num, spec, sizeof(spec) / sizeof(spec[0]));
+	}
+	
+}
+
 static void ready_input(ErlDrvData e, ErlDrvEvent ready_fd)
 {
     int fd = (int)(long)e;
@@ -2184,8 +2253,8 @@ static void ready_input(ErlDrvData e, ErlDrvEvent ready_fd)
 	}
 	else if (res == 0)
 	    port_inp_failure(port_num, ready_fd, res);
-	else 
-	    driver_output(port_num, (char*) read_buf, res);
+	else
+		send_from_spawned_process(fd, (char *) read_buf, res);
 	erts_free(ERTS_ALC_T_SYS_READ_BUF, (void *) read_buf);
     }
     else if (fd_data[ready_fd].remain > 0) { /* We try to read the remainder */
@@ -2200,7 +2269,7 @@ static void ready_input(ErlDrvData e, ErlDrvEvent ready_fd)
 	    port_inp_failure(port_num, ready_fd, res);
 	}
 	else if (res == fd_data[ready_fd].remain) { /* we're done  */
-	    driver_output(port_num, fd_data[ready_fd].buf, 
+	    send_from_spawned_process(fd, fd_data[ready_fd].buf, 
 			  fd_data[ready_fd].sz);
 	    clear_fd_data(ready_fd);
 	}
@@ -2254,7 +2323,7 @@ static void ready_input(ErlDrvData e, ErlDrvEvent ready_fd)
 		}
 
 		if (h <= (bytes_left)) {
-		    driver_output(port_num, (char*) cpos, h);
+		    send_from_spawned_process(fd, (char*) cpos, h);
 		    cpos += h;
 		    bytes_left -= h;
 		    continue;
@@ -2703,6 +2772,14 @@ report_exit_status(ErtsSysReportExit *rep, int status)
 				 (ERL_DRV_WRITE|ERL_DRV_USE),
 				 1);
 	}
+	if (rep->efd >= 0) {
+	    driver_data[rep->efd].alive = 0;
+	    driver_data[rep->efd].status = status;
+	    (void) driver_select((ErlDrvPort) internal_port_index(pp->id),
+				 rep->efd,
+				 (ERL_DRV_READ|ERL_DRV_USE),
+				 1);
+	}
 	erts_port_release(pp);
     }
     erts_free(ERTS_ALC_T_PRT_REP_EXIT, rep);
-- 
1.6.4.2

